/*
 * Create image with sparse values.  
 *
 * Each value should be preceded with 2D or 3D coordinate according to
 * image format.
 *  
 * See also: cim (full values)
 * 
 * (c) 2017 --  D.Béréziat
 */


#include <inrimage/image.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern int debug_;

int getint( char **str, int *val) {
  char svalue[20], *s;
  if( sscanf( *str, "%s", svalue) <= 0)
    return 0;
  *val = atoi(svalue);
  *str = strstr( *str, svalue) + strlen(svalue);
  return 1;
}

int getfloat( char **str, float *val) {
  char svalue[40], *s;
  if( sscanf( *str, "%s", svalue) <= 0)
    return 0;
  *val = atof(svalue);
  *str = strstr( *str, svalue) + strlen(svalue);
  return 1;
}


int main(int argc, char **argv) {
  struct image *nf;
  Fort_int lfmt[9];

  float *fdata;
  char *idata;
  
  char name [512];
  int x, y, z;
  int type_image;
  int exist, writable;
  int current_line, l;
  int b1;
  char format[20] = "%d";
  int readint, bsize;

  Fort_int icodi[2] = {1,FIXE}, icodo[2] = {sizeof(float), REELLE}, iexpi=200, iexpo=0;
  
  inr_init(argc, argv, "1.0", "[image|options format] [-b0][-f=fmt]",
	   "Read from stdin sparse ascii values to input in an image.\n"
	   "Each line read should have the following format : x y [z] value [value ...]\n"
	   "where x y z are pixel coordinates and value is pixel value (several values for an image vector).\n"
	   "Options are:\n"
	   "\t-b0: indices start from 0 (1 default)\n"
	   "\t-f=fmt: force format for reading values from stdin with scanf.\n"
	   "\t        Possible formats are: %d, %x, %c, %f and %g\n");
	     
  outfileopt(name);
  type_image = is_inrimage(name,1);
  exist = type_image>0;
  writable = type_image<IV_MAX;
  nf = image_( name, "s", "", lfmt);

  /* indices start to 0 */
  b1 = 1-igetopt0("-b0");
  if( debug_ && !b1) fprintf(stderr, "*indices start to 0\n");

  /* fix input format */
  if( igetopt0x("-f=","%s",format) ) { 
    switch (format[1]) {
    case 'd':
    case 'x':
    case 'c':
      readint = 1;
      break;
    case 'f':
    case 'g':
      readint = 0;
      break;
    default:
      imerror(9,"Error with option -f=%s, only format %%d, %%x, %%c, %%f or %%g are supported.\n", format);
    }
  } else {
    /* format of input value are determined from image coding */
    if( TYPE!=REELLE && (BSIZE == 1 || (BSIZE<0 && BSIZE>-8)) ) {
      readint = 1;
    } else {
      readint = 0;
      format[1] = 'f';
    }
  }

  iexpi = EXP;  

  if(debug_) {
    int i;
    fprintf(stderr, "fmt of output image: "); 
    for(i=0;i<9;i++) fprintf(stderr, "%d ", lfmt[i]);
    fprintf(stderr,"\nreadint=%d, format=%s\n", readint, format);
    fprintf(stderr,"icodi: %d %d %d icodo: %d %d %d\n", icodi[0], icodi[1], iexpi, icodo[0], icodo[1], iexpo);
  }
  
  if( exist) {
    if( writable) {
      /* image exists, allocate one line */
      fdata = (float*)i_malloc(sizeof(float)*DIMX);
      if (readint) idata = (char*)i_malloc(sizeof(char)*DIMX);
    } else
      imerror(13, "Image exists but can not be written!\n");
  } else {
    fdata = (void*)i_calloc(sizeof(float),DIMX*DIMY);
    if (readint) idata = (char*)i_calloc(sizeof(char), DIMX*DIMY);
  }

  z=0;
  current_line = -1;
  l=1;
  while (fgets(name, 511, stdin)) {
    char *pname = name;
    
    if( !(getint( &pname, &x) && getint( &pname, &y)) ) 
      imerror(9, "Bad format at line %d (expected: x y values)\n", l);
    if (NDIMZ > 1) {
      if( !getint( &pname, &z)) 
	imerror(9, "Bad format at line %d (expected: x y z values)\n", l);
      z -= b1;
    }

    if( debug_) fprintf(stderr, "read (%d,%d,%d) %s\n", x, y, z+b1, pname);
    x -= b1;
    y -= b1;
        
    if( x<NDIMX && y<NDIMY && z<NDIMZ && x>=0 && y>=0 && z>=0) {
      long idx;
      int v;

      if( exist && writable && current_line != y + NDIMY*z) {
	if( debug_) fprintf(stderr, "current line: %d new line: %d\n",
			    current_line,y+x*NDIMY);
	current_line = y + NDIMY*z;
	c_lptset( nf, 1+current_line);
	c_lecflt( nf, 1, fdata);
	if(readint)  c_cnvtbg((char*)fdata,idata,DIMX,icodo,icodi,iexpo,iexpi);
      }
      
      if(  exist && writable) {
	idx = NDIMV*x;
      } else
	idx = NDIMV*(x+NDIMX*(y+NDIMY*z));

      if(debug_) fprintf(stderr,"writing values at offset %ld\n", idx);
      for( v=0; v < NDIMV; v++) {
	int res, val;
	if (readint) {
	  res = getint( &pname, &val);
	  idata[idx+v] = val;
	} else
	  res  = getfloat( &pname, fdata+idx+v);
	if ( !res) imerror(9, "read value error: missing a value at line %d\n", l);
      }
    } else      
      fprintf(stderr,
	      NDIMZ==1?
	      "Warning line %d: coord (%d,%d) outside the domain, values ignored\n":
	      "Warning line %d: coord (%d,%d,%d) outside the domain, values ignored\n",
	      l, x+b1, y+b1, z+b1);

    if( exist && writable) {
      if( readint) c_cnvtbg(idata,(char*)fdata,DIMX,icodi,icodo,iexpi,iexpo);      
      c_lptset( nf, 1+current_line);
      c_ecrflt( nf, 1, fdata);
    }

    l++;
  }

  if( !(exist && writable)) {
    if( readint) c_cnvtbg(idata,(char*)fdata,DIMX*DIMY,icodi,icodo,iexpi,iexpo);  
    c_ecrflt( nf, DIMY, fdata);
  }
  fermnf_(&nf);
  return 0;
}
